require 'sinatra'
require 'json'
require 'logger'
require 'asm'
require 'asm/data/deployment'
require 'asm/service_deployment'
require 'asm/device_management'
require 'asm/appliance_setup/dhcp'
require 'asm/errors'
require 'asm/monitoring'

class ASM::App < Sinatra::Base

  configure do
    set :bind, '0.0.0.0'
    # only allow a single request to be processed at a time
    set :lock, true

    ASM.init
    ASM.logger.info('ASM deployment service initialized')

    # Since we have just started there can be no in-progress deployments
    ASM::Data::Deployment.mark_in_progress_failed(ASM.database, ASM.logger)
  end

  # Initiate migration of server
  post '/process_service_profile_migration' do
    content_type :json
    ASM.process_deployment_migration(request).to_json
  end

  get '/puppetreport/:id/:certname' do |id, certname|
    content_type :json
    report = ASM::Util.get_report(id, certname)
    report.to_json
  end

  get '/puppetlog/:id/:certname' do |id, certname|
    content_type :json
    log = ASM::Util.get_puppet_log(id, certname)
    log.to_json
  end

  get '/status' do
    content_type :json

    status = {"active_deployments" => ASM.active_deployments,
              "running_certnames" => ASM.running_certnames,
              "large_process_count" => ASM.get_counter(:large_child_procs)}

    status.to_json
  end

  # Execute deployment
  post '/deployment' do
    content_type :json
    ASM.process_deployment_request(request).to_json
  end

  get '/deployment/:id/status' do |id|
    begin
      content_type :json
      ASM.get_deployment_status(id).to_json
    rescue ASM::NotFoundException
      status 404
    end
  end

  get '/deployment/:id/log' do |id|
    begin
      content_type :json
      ASM.get_deployment_log(id).to_json
    rescue ASM::NotFoundException
      status 404
    end
  end

  get '/deployment/:id/puppet_logs/:certname' do |id,certname|
    begin
      content_type :json
      ASM::Util.get_puppet_component_run(id,certname)
    rescue Errno::ENOENT
      status 404
    end
  end

  put '/deployment/:id' do |id|
    begin
      content_type :json
      ASM.retry_deployment(id, JSON.parse(request.body.read)).to_json
    rescue ASM::NotFoundException
      status 404
    end
  end

  delete '/deployment/:id' do | id |
    begin
      ASM.clean_deployment(id)
    rescue ASM::NotFoundException
      status 404
    end
  end

  delete '/device/:cert_name' do |cert_name|
    ASM.execute_async(ASM.logger) do
      ASM::DeviceManagement.remove_device(cert_name)
    end
  end

  post '/device' do
    content_type :json

    device = JSON.parse(request.body.read)

    begin
      ASM::DeviceManagement.write_device_config(device)

      ASM.execute_async(ASM.logger) do
        ASM::DeviceManagement.run_puppet_device!(device["cert_name"], ASM.logger)
      end

      status 202
      ASM::DeviceManagement.get_device(device["cert_name"]).to_json
    rescue ASM::DeviceManagement::DuplicateDevice
      status 409
    end
  end

  get '/device/:cert_name' do |cert_name|
    content_type :json

    begin
      ASM::DeviceManagement.get_device(cert_name).to_json
    rescue ASM::NotFoundException
      status 404
    end
  end

  get '/metrics/:ref_id/?:from?/?:units?' do |ref_id, from, units|
    begin
      content_type :json
      required = (params[:required] || '').split(',')
      ASM.get_metrics(ref_id, from, units, required)
    rescue ASM::NotFoundException
      status 404
    end
  end

  put '/device/:cert_name' do |cert_name|
    content_type :json

    device = JSON.parse(request.body.read)
    ASM::DeviceManagement.write_device_config!(device)

    ASM.execute_async(ASM.logger) do
      ASM::DeviceManagement.run_puppet_device!(cert_name, ASM.logger)
    end

    status 202
    ASM::DeviceManagement.get_device(cert_name).to_json
  end

  put '/dhcp' do
    result = ASM.process_dhcp_request(request)
    status result['status']
    result.to_json
  end

  put '/process_monitoring_data' do
    ASM.process_monitoring_data(request)
    status 200
  end
end
