require 'asm/device_management'
require 'asm/ipmi'

class ASM::ServiceDeployment::Teardown
  attr_reader :logger

  def initialize(service_deployment, logger)
    @service_deployment = service_deployment
    @logger = logger
    @debug = service_deployment.debug?
  end

  def log(msg)
    logger.info(msg)
  end

  def thread_with_exception_log(component_name)
    Thread.new do
      begin
        yield
      rescue Exception
        Thread.current[:exception] = $!
        @service_deployment.write_exception(component_name, $!)
      end
    end
  end

  def process_virtualmachine_teardown_component(component)
    thread_with_exception_log(component["puppetCertName"]) do
      # code copied verbatim from ASM::ServiceDeployment#process_virtualmachine and should be DRY'd up
      resource_hash = ASM::Util.build_component_configuration(component, :decrypt => @service_deployment.decrypt?)

      vms = ASM::Resource::VM.create(resource_hash)
      raise(Exception, "Expect one set of VM configuration #{vm.size} configuration recieved.") unless vms.size == 1
      vm = vms.first
      vm.ensure = "absent"

      servers = ASM::Resource::Server.create(resource_hash)
      raise(Exception, "Expect zero or one set of Server configuration: #{servers.size} were passed") if servers.size > 1
      server = servers.first

      clusters = (@service_deployment.find_related_components('CLUSTER', component) || [])
      cluster = clusters.first || {}
      cluster_certname = cluster['puppetCertName']
      raise(Exception, "Expect one cluster for #{cluster_certname}: #{clusters.size} was passed") unless clusters.size == 1

      cluster_deviceconf = ASM::DeviceManagement.parse_device_config(cluster_certname)
      cluster_resource = ASM::Util.build_component_configuration(cluster, :decrypt => @service_deployment.decrypt?)
      clusters = ASM::Resource::Cluster.create(cluster_resource)
      raise(Exception, "Expected one asm::cluster resource: #{clusters.size} was provided") unless clusters.size == 1
      title, cluster_params = clusters.first.shift
      cluster_params.title = title

      vm.process!(cluster_certname, server, cluster_params,@service_deployment.id, @logger)

      hostname = vm.hostname || server.os_host_name
      raise(ArgumentError, 'VM hostname not specified, missing server os_host_name value') unless hostname

      resource_hash = vm.to_puppet
      vm_resources = resource_hash[resource_hash.keys[0]]
      vm_title = vm_resources.keys[0]
      vm_resource = vm_resources[vm_title]

      # vm.process! sets a requested_network_interfaces parameter that is not
      # supported by asm::vmware::vcenter. network_interfaces is not needed either
      # for ensure = absent
      vm_resource.delete('network_interfaces')
      vm_resource.delete('requested_network_interfaces')

      if server
        serial_number = ''
        server.process!(serial_number, @service_deployment.id)
        server.title = vm_title
        server.ensure = 'absent'
        resource_hash['asm::server'] = server.to_puppet
      end

      certname = "vm-%s" % hostname.downcase
      resource_hash.delete('asm::vm::vcenter') if !component['teardown']
      resource_hash.delete('asm::vm::scvmm') if !component['teardown']
      @service_deployment.process_generic(certname, resource_hash, 'apply')
    end
  end

  def process_server_teardown_component(server)
    thread_with_exception_log(server["puppetCertName"]) do
      cert_name = server['puppetCertName'] or raise(Exception, 'Component has no certname')
      server_conf = ASM::Util.build_component_configuration(server, :decrypt => @service_deployment.decrypt?)
      # Check if the server has a base::server, then we need the cleanup of backup server
      oldserver_info = (server_conf['asm::baseserver'] || {})
      if !oldserver_info.empty?
        cert_name = oldserver_info.keys[0]
      end

      server_resource = ASM::Resource::Server.create(server_conf).first

      if server_resource && !server_resource.os_image_type.nil?
        # NOTE: It is critical that the razor policies get removed from the
        # server or else it will not be able to be used in any future
        # deployment. So execute just that bit here before everything else.

        serial_number = ASM::Util.cert2serial(cert_name)
        server_resource.process!(serial_number, @service_deployment.id)
        server_resource.ensure = 'absent'
        resource_hash = { 'asm::server' => server_resource.to_puppet }
        # Clean the certificate name created for the old server
        log("Deleting certificate for server #{server_resource.os_host_name}")
        ASM::DeviceManagement.clean_cert(ASM::Util.hostname_to_certname(server_resource.os_host_name))

        @service_deployment.process_generic(cert_name, resource_hash, 'apply', true)
      end

      idrac = server['resources'].detect{|r|r['id']=='asm::idrac'}

      if idrac
        log("Cleanup associated storage after teardown of #{cert_name}")
        related_components = @service_deployment.find_related_components('STORAGE',server)
        related_components.each do |component|
          storage_hash = ASM::Util.build_component_configuration(component, :decrypt => @service_deployment.decrypt?)
          if storage_hash['asm::volume::equallogic']
            unless component['teardown']
              log("Cleanup associated equallogic storage #{component['name']} after removal of #{cert_name}")
              @service_deployment.process_storage(component)
            end
          end
        end

        log("Cleaning virtual identities from #{cert_name}")
        begin
          endpoint = ASM::DeviceManagement.parse_device_config(cert_name)
          @service_deployment.cleanup_server(server, cert_name)
          ASM::WsMan.wait_for_lc_ready(endpoint, @logger) unless @debug
        ensure
          cleanup_compellent_serverobject(server)
          ASM::WsMan.poweroff(endpoint, @logger) unless @debug
        end
      else
        cert_name = server['puppetCertName']
        endpoint = ASM::DeviceManagement.parse_device_config(cert_name)
        is_dell_server = ASM::Util.dell_cert?(cert_name)
        if is_dell_server
          ASM::WsMan.poweroff(endpoint, @logger) unless @debug
        else
          ASM::Ipmi.power_off(endpoint, @logger) unless @debug
        end
      end
    end
  end

  def cleanup_compellent_serverobject(server)
    # Delete the compellent server object when storage object is not selected for deletion.

    # Check if the server has compellent as related storage object and its teardown
    # flag is false. If false then just delete the server object
    related_storage_components = @service_deployment.find_related_components('STORAGE',server)
    service_tag = ASM::Util.cert2serial(server['puppetCertName'])
    server_object_name = "ASM_#{service_tag}"
    related_storage_components.each do |storage_component|
      next if !storage_component['puppetCertName'].match(/compellent/)
      server_folder = (storage_component['server_folder'] || '')
      compellent_skipped = related_storage_components.find_all {
          |x| !x['teardown']
      }
      @logger.debug("Compellent components that needs to be skipped: #{compellent_skipped}")
      (compellent_skipped || [] ).each do |compellent|
        new_resource_hash = {}
        new_resource_hash['compellent_server'] ||= {}
        new_resource_hash['compellent_server'][server_object_name] ||= {}
        new_resource_hash['compellent_server'][server_object_name] = {
            'ensure'          => 'absent',
            'serverfolder'    => server_folder,
        }
        @logger.debug("New resource hash for compellent storage group: #{new_resource_hash}")
        begin
          @service_deployment.process_generic(compellent['puppetCertName'],
                                              new_resource_hash, 'device', true, nil, compellent['asmGUID'])
        rescue Exception => e
          @logger.debug("Failed to cleanup the server object #{server_object_name}")
          @logger.debug("Error: #{e.inspect}")
        end
        # Clean-up of the first object is enough as server object is common
        break
      end
      # Cleanup for first volume component will be enough
      @logger.debug("Completed cleanup for the first compellent object")
      break
    end
  end

  def process_vmware_cluster_teardown(cert_name, component, resource_hash)
    cluster_hash = ASM::Util.build_component_configuration(component, :decrypt => @service_deployment.decrypt?)
    cluster_title = cluster_hash['asm::cluster'].keys[0]
    cluster_params = (cluster_hash['asm::cluster'] || {})[cluster_title]

    #Want to remove the hosts separately, in case the cluster itself isn't being torn down.
    related_servers = @service_deployment.find_related_components('SERVER', component)
    related_servers.each do |server|
      if server['teardown']
        server_cert = server['puppetCertName']
        server_hash = ASM::Util.build_component_configuration(server, :decrypt => @service_deployment.decrypt?)
        server_params = (server_hash['asm::server'] || {})[server_cert]
        network_params = (server_hash['asm::esxiscsiconfig'] || {})[server_cert]
        if network_params
          net_config = ASM::NetworkConfiguration.new(network_params['network_configuration'])
          begin
            mgmt_network = net_config.get_network('HYPERVISOR_MANAGEMENT')
            static = mgmt_network['staticNetworkConfiguration']
            hostname = @service_deployment.lookup_hostname(server_params['os_host_name'], static).downcase
          rescue Exception => e
            @logger.debug("There is no management network associated with the server: #{server_cert}")
            @logger.debug("Error message: #{e.inspect}")
          end
          teardown_conf = {'asm::host' => {},
                           'transport' => {
                               'vcenter' => {
                                   'name' => cluster_title,
                                   'options' => {'insecure' => true},
                                   'provider' => 'device_file'}}}
          teardown_conf['asm::host'][server_cert] = {
              'datacenter' => cluster_params['datacenter'],
              'cluster' => cluster_params['cluster'],
              'hostname' => hostname,
              'username' => ASM::ServiceDeployment::ESXI_ADMIN_USER,
              'password' => server_params['admin_password'],
              'decrypt' => @service_deployment.decrypt?,
              'timeout' => 90,
              'ensure' => 'absent'
          }
          @logger.info("Removing #{server_cert} from cluster.")
          @service_deployment.process_generic(server_cert, teardown_conf, apply_type(server['type']), true)
        end
      end
    end
    if component['teardown']
      (resource_hash['asm::cluster'] || {}).each do |title, params|
        resource_hash['asm::cluster'][title]['vcenter_options'] = { 'insecure' => true }
        resource_hash['asm::cluster'][title]['ensure'] = 'absent'
        resource_hash['asm::cluster'][title].delete('ha_config')
        resource_hash['asm::cluster'][title].delete('drs_config')
        resource_hash['asm::cluster'][title].delete('vsan')

        @service_deployment.process_generic(cert_name, resource_hash, apply_type(component['type']), true)
      end
    end
  end

  def process_scvmm_cluster_teardown(cert_name, component, resource_hash)
    (resource_hash['asm::cluster::scvmm'] || {}).each do |title,params|
      log("Tearing down SCVMM cluster %s" % title)
      @service_deployment.configure_hyperv_cluster(component, resource_hash, title, "absent")
    end
  end

  def process_cluster_teardown_component(component)
    # need to do what ASM::ServiceDeployment#process_cluster does in reverse but that
    # method does a really huge amount of work and I suspect on teardown we can just
    # delete the hosts and cluster which will remove the rest of the stuff
    #
    # there are some issues in vc_host where it auto requires the cluster - which on
    # teardown means the order is always wrong, so relying on vmware to clean up all
    # hosts/vms/etc in a cluster when we just delete the cluster
    thread_with_exception_log(component["puppetCertName"]) do
      cert_name = component['puppetCertName']
      raise(Exception, 'Component has no certname') unless cert_name

      log("Tearing down cluster component: %s" % cert_name)

      resource_hash = ASM::Util.build_component_configuration(component, :decrypt => @service_deployment.decrypt?)

      if resource_hash.include?("asm::cluster")
        process_vmware_cluster_teardown(cert_name, component, resource_hash)
      elsif resource_hash.include?("asm::cluster::scvmm")
        process_scvmm_cluster_teardown(cert_name, component, resource_hash)
      else
        raise("Do not know how to tear down cluster '%s (%s)'" % [component["name"], component["puppetCertName"]])
      end
    end
  end

  def remove_storage_component_from_cluster(component, resource_hash)
    related_servers = @service_deployment.find_related_components("SERVER", component)

    unless related_servers.empty?
      related_server = related_servers.first

      related_cluster = @service_deployment.find_related_components("CLUSTER", related_server)

      unless related_cluster.empty?
        storage_cert = component["puppetCertName"]
        storage_creds = ASM::DeviceManagement.parse_device_config(storage_cert)

        related_cluster = related_cluster.first
        cluster_cert = related_cluster["puppetCertName"]

        server_conf = ASM::Util.build_component_configuration(related_server, :decrypt => true)
        server_cert = related_server["puppetCertName"]

        network_params = (server_conf['asm::esxiscsiconfig'] || {})[server_cert]
        network_config = ASM::NetworkConfiguration.new(network_params['network_configuration'], @logger)
        mgmt_network = network_config.get_network('HYPERVISOR_MANAGEMENT')
        static = mgmt_network['staticNetworkConfiguration']
        hostip = static['ipAddress']

        puppet_hash = {"transport" => {"vcenter" => {"name" => cluster_cert,
                                                     "options" => {"insecure" => true},
                                                     "provider" => "device_file"}}}

        if resource_hash["asm::volume::equallogic"]
          resource_hash["asm::volume::equallogic"].each do |storage_title, storage_params|
            if @debug
              target_iqn = "DEBUG-IQN-#{storage_title}"
            else
              target_iqn = ASM::Util.get_eql_volume_iqn(component['asmGUID'], storage_title)
            end

            puppet_hash['esx_datastore'] = {"#{hostip}:#{storage_title}" => {"ensure" => "absent",
                                                                             "datastore" => storage_title,
                                                                             "type" => "vmfs",
                                                                             "target_iqn" => target_iqn,
                                                                             "transport" => "Transport[vcenter]"}}
          end
        elsif resource_hash['asm::volume::compellent']
          resource_hash['asm::volume::compellent'].each do |storage_title, storage_params|
            if @debug
              lun_id = 0
            else
              device_id = ASM::Util.find_compellent_volume_info(component['asmGUID'], storage_title, storage_params["volumefolder"], @logger)
              decrypt_password = ASM::Cipher.decrypt_string(server_conf["asm::server"][server_cert]["admin_password"])
              lun_id = @service_deployment.get_compellent_lunid(hostip, 'root', decrypt_password, device_id)
            end

            puppet_hash['esx_datastore'] = {"#{hostip}:#{storage_title}" => {"ensure" => "absent",
                                                                             "datastore" => storage_title,
                                                                             "type" => "vmfs",
                                                                             "lun" => lun_id,
                                                                             "transport" => "Transport[vcenter]"}}
          end
        elsif resource_hash['netapp::create_nfs_export']
          resource_hash['netapp::create_nfs_export'].each do |storage_title, storage_params|
            dc = related_cluster["datacenter"]
            cluster = related_cluster["cluster"]

            puppet_hash['esx_datastore'] = {"#{hostip}:#{storage_title}" => {"ensure" => "absent",
                                                                             "path" => "/#{dc}/#{cluster}/",
                                                                             "type" => "nfs",
                                                                             "transport" => "Transport[vcenter]"}}
          end
        else
          raise("Could not find any supported storage type to remove from the cluster for component %s" % component['puppetCertName'])
        end

        @service_deployment.process_generic(component['puppetCertName'], puppet_hash, 'apply', true, nil, component['asmGUID'])
      end
    end
  rescue Exception => e
    @logger.debug("Could not remove storage component %s from any cluster: %s" % [component['puppetCertName'], e.to_s])
    raise e
  end

  def cleanup_compellent_storage_volumes(component, resource_hash)
    process_storage = true

    (resource_hash['asm::volume::compellent'] || {}).each do |title, params|
      log("Tearing down compellent storage %s" % title)
      related_servers = @service_deployment.find_related_components('SERVER', component)
      resource_hash['asm::volume::compellent'][title].delete('configuresan')
      resource_hash['asm::volume::compellent'][title]['force'] = 'true'
      resource_hash['asm::volume::compellent'][title]['manual'] = 'true'

      # Standalone storage component is passed
      # Just create the volume
      if related_servers.empty?
        resource_hash['asm::volume::compellent'][title]['servername'] = ""
        @service_deployment.process_generic(component['puppetCertName'], resource_hash, 'device', true, nil, component['asmGUID'])
      end

      related_servers.each do |server_comp|
        server_servicetag = ASM::Util.cert2serial(server_comp['puppetCertName'])
        server_teardown_flag = server_comp['teardown']
        resource_hash['asm::volume::compellent'][title]['servername'] = "ASM_#{server_servicetag}" if server_teardown_flag

        @service_deployment.process_generic(component['puppetCertName'], resource_hash, 'device', true, nil, component['asmGUID'])
      end
      process_storage = false
    end

    process_storage
  end

  def process_storage_teardown_component(component)
    thread_with_exception_log(component["puppetCertName"]) do
      resource_hash = ASM::Util.build_component_configuration(component, :decrypt => true)

      begin
        remove_storage_component_from_cluster(component, resource_hash)
      rescue Exception => e
        @logger.debug("Could not remove the storage association from the cluster, continuing to tear down volumes")
      end

      process_storage = cleanup_compellent_storage_volumes(component, resource_hash)

      @service_deployment.process_generic(component['puppetCertName'], resource_hash, 'device', true, nil, component['asmGUID']) if process_storage
    end
  end

  def apply_type(device_type)
    device_type.downcase == 'storage' ? 'device' : 'apply'
  end

  def process_teardown_component(component)
    thread_with_exception_log(component["puppetCertName"]) do
      raise("Component has no certname") unless component["puppetCertName"]

      teardown_resources = ASM::Util.build_component_configuration(component, :decrypt => @service_deployment.decrypt?)

      log("Tearing down component '%s (%s)' of type '%s' using %d resource(s)" % [component["name"], component["puppetCertName"], component["type"], teardown_resources.size])

      @service_deployment.process_generic(component["puppetCertName"], teardown_resources, apply_type(component['type']), true)
    end
  end

  def teardown_method_for(type)
    "process_#{type.downcase}_teardown_component".intern
  end

  def has_teardown_method_for?(type)
    respond_to?(teardown_method_for(type))
  end

  def process_teardown(service_deployment)
    log("Tearing down deployment %s with id %s" % [service_deployment['deploymentName'], service_deployment['id']])

    exceptions = []

    ['TEST', 'VIRTUALMACHINE', 'CLUSTER', 'SERVER', 'STORAGE'].each do |type|
      teardown_candidates = service_deployment["serviceTemplate"]["components"].select do |c|
        (c['teardown'] || (c['type'] == 'CLUSTER' || c['type'] == 'VIRTUALMACHINE')) && c['type'] == type
      end

      unless teardown_candidates.empty?
        threads = []

        teardown_candidates.each do |component|
          if has_teardown_method_for?(type)
            threads << send(teardown_method_for(type), component)
          else
            threads << process_teardown_component(component)
          end
        end

        threads.each do |thread|
          thread.join
          exceptions << thread[:exception] if thread[:exception]
        end
      end
    end

    # doing this here ensures that we attempt to tear down as much as possible
    # rather than bail on the first swim lane that fails, all swim lanes run
    # all exceptions are logged from each swim lane and the first failure is
    # raised, bit of a pity we lose the other exceptions but this pattern was
    # used elsewhere already
    unless exceptions.empty?
      exceptions.each do |e|
        log("Exception during teardown: %s (%s)" % [e.inspect, e.class])
      end

      raise exceptions.first
    end
  end
end
