#!/opt/puppet/bin/ruby

require 'asm'
require 'asm/cipher'
require 'erb'

def conf
  @config ||= database_config(ASM::Util::DATABASE_CONF)
end

def connect_db(db_name)
  if RUBY_PLATFORM == 'java'
    require 'jdbc/postgres'
    Jdbc::Postgres.load_driver
    Sequel.connect("jdbc:postgresql://#{conf.host}/#{db_name}?user=#{conf.username}&password=#{conf.password}")
  else
    require 'pg'
    Sequel.connect("postgres://#{conf.username}:#{conf.password}@#{conf.host}:#{conf.port}/#{db_name}")
  end
end

def inventory_db
  @inventory_db ||= connect_db("asm_manager")
end

def credentials_db
  @credentials_db ||= connect_db("encryptionmgr")
end

def chassis_db
  @credentials_db ||= connect_db("chassis_ra")
end

def get_credentials(cred_id)
  # should be able to just use ASM::Cipher#decrypt_credential but this is
  # failing for cases where any password id is nil and dont want to yak shave
  # that now
  creds = ASM::Cipher.db[ASM::Cipher.credential_query, cred_id].first
  creds.clone.each do |key, val|
    if key.to_s =~ /^(.+?)_?id$/ && val
      creds[$1.to_sym] = ASM::Cipher.decrypt_string(val)
    end
  end

  creds
end

def path_to_template(template)
  File.read(File.expand_path(File.join(File.dirname(__FILE__), "templates", template)))
end

def render_host(server, chassis)
  ERB.new(path_to_template("host.erb")).result(binding)
end

def render_services(server, chassis)
  erb = path_to_template("check_%s_hardware_health.erb" % server[:device_type])
  ERB.new(erb).result(binding)
end

def render_related_os_services(server, chassis)
  erb = path_to_template("check_%s_os_health.erb" % server[:os_image_type])
  ERB.new(erb).result(binding)
end

def chassis_for_switch(svctag, resources)
  chassis = chassis_db["SELECT chassis.*, iom.slot FROM chassis, iom  WHERE chassis.id = iom.chassis_ref_id AND iom.service_tag='#{svctag}'"]

  result = nil

  if chassis.count == 1
    if chassis_tag = chassis.first[:service_tag]
      matched_chassis = resources.select{|r| r[:service_tag] == chassis_tag}

      if matched_chassis.count == 1
      	result = matched_chassis.first.to_hash
	result[:chassis_slot] = chassis.first[:slot]
      end
    end
  end

  result
end

resources = inventory_db["SELECT * FROM device_inventory WHERE device_type in ('BladeServer', 'ChassisM1000e', 'RackServer', 'dellswitch', 'Server', 'ChassisFX', 'FXServer') ORDER BY ref_id ASC"].to_a

config_file = [File.read(File.expand_path(File.join(File.dirname(__FILE__), "templates", "commands.cfg")))]

resources.each do |server|
  server[:credentials] = get_credentials(server[:cred_id])
  server[:ip_address] = server[:ip_address].to_s
end

resources.each do |server|
  chassis = {}

  if server[:device_type] == "dellswitch"
    # skip all but switches in chassis
    next unless chassis = chassis_for_switch(server[:service_tag], resources)
  end

  config_file << render_host(server, chassis)
  config_file << render_services(server, chassis)

  # disk checks via the OS is only done on C-Series machines, device type "Server"
  if (server[:model].match(/^PowerEdge C\d+/) && server[:os_ip_address] && !server[:os_ip_address].empty?)
    config_file << render_related_os_services(server, chassis)
  end
end

new_config = config_file.join("\n")
old_config = File.read("/etc/nagios/conf.d/asm.cfg") rescue ""

unless new_config == old_config
  File.open("/etc/nagios/conf.d/asm.cfg", "w") do |f|
    f.puts new_config
  end

  system("/sbin/service nagios reload 2>&1 >/dev/null")
end
