require 'asm/service_deployment'
require 'asm/service_deployment/teardown'
require 'json'
require 'spec_helper'
require 'yaml'
require 'asm/util'

describe ASM::ServiceDeployment::Teardown do

  before do
    ASM.init_for_tests
  end

  after do
    ASM.reset
  end

  describe "when tearing down" do
    before(:each) do
      @logger = mock
      @logger.stubs(:debug?).returns(false)
      @logger.stubs(:debug)
      @logger.stubs(:info)

      @sd = mock
      @sd.stubs(:decrypt?).returns(true)
      @sd.stubs(:debug?).returns(false)
      @sd.stubs(:id).returns('8888')
      @sd.stubs(:write_exception)

      @td = ASM::ServiceDeployment::Teardown.new(@sd, @logger)

      deployment_file = File.join(File.dirname(__FILE__), '..', '..', '..', 'fixtures', 'service_deployment_teardown_test.json')
      @service_deployment = JSON.parse(File.read(deployment_file), :max_nesting => 100)

      @mock_thread = mock
      @mock_thread.stubs(:join)
      @mock_thread.stubs(:[]).returns(nil)

      @components = @service_deployment["serviceTemplate"]["components"]
      @teardown_candidates = @components.select{|c| ["STORAGE", "VIRTUALMACHINE", "CLUSTER", "SERVER"].include?(c["type"])}
      @td.stubs(:cleanup_compellent_serverobject).returns(true)
    end

    it "should correctly determine teardown method names" do
      @td.teardown_method_for("RSPEC").should == :process_rspec_teardown_component
    end

    it "should process each component with the correct handler" do
      ["STORAGE", "VIRTUALMACHINE", "CLUSTER", "SERVER"].each do |type|
        @components.select{|c| c["type"] == type}.each do |component|
          @td.expects(@td.teardown_method_for(type)).with(component).returns(@mock_thread)
        end
      end

      @td.process_teardown(@service_deployment)
    end

    it "should capture and raise thread exceptions" do
      e = ASM::Error.new("rspec test")
      e.set_backtrace(["rspec"])

      @mock_thread.stubs(:[]).returns(e)

      @td.stubs(:has_teardown_method_for?).returns(false)
      @td.stubs(:process_teardown_component).returns(@mock_thread)

      @td.expects(:log).with(regexp_matches /Tearing down deployment test service/)
      @td.expects(:log).with("Exception during teardown: #<ASM::Error: rspec test> (ASM::Error)").times(5)

      expect { @td.process_teardown(@service_deployment) }.to raise_error("rspec test")
    end

    describe "when doing server teardowns" do
      it "should check the component has a certname" do
        thread = @td.process_server_teardown_component({"resources" => [{"id" => "asm::idrac"}]}).join
        thread[:exception].message.should == "Component has no certname"
      end

      it "should only teardown when there are asm::idrac resources" do
        @ipmi = ASM::Ipmi
        @ipmi.stubs(:power_off).returns(true)
        @td.expects(:log).with(regexp_matches /Cleaning virtual identities from/).never

        thread = @td.process_server_teardown_component({'puppetCertName' => 'test-cert',
                                                        'resources' => []}).join

        raise thread[:exception] if thread[:exception]
      end

      it "should clean the server using the correct cert name and power it off" do
        component = @components.select{|c| c["type"] == "SERVER"}.first

        ASM::DeviceManagement.expects(:parse_device_config).returns("rspec")
        ASM::DeviceManagement.expects(:clean_cert)
        ASM::WsMan.expects(:wait_for_lc_ready).with("rspec", @logger)
        ASM::WsMan.expects(:poweroff).with("rspec", @logger)
        expected_cert = 'bladeserver-jgd5hw1'
        @sd.expects(:cleanup_server).with(component, expected_cert)
        @sd.expects(:find_related_components).returns([])

        # Expect razor policies to be removed
        teardown_hash = {
            'asm::server' => {expected_cert => {
              'os_host_name' => 'arri-testhost',
              'razor_image' => 'esxi-5.1',
              'admin_password' => 'ff80808148a06b000148a6d3e6f70acf',
              'decrypt' => true,
              'esx_mem' => 'false',
              'ensure' => 'absent',
              'broker_type' => 'noop',
              'serial_number' => 'JGD5HW1',
              'policy_name' => 'policy-arri-testhost-8888',
              'installer_options' => {'os_type' => 'vmware_esxi', 'agent_certname' => 'agent-arri-testhost'}
            }}
        }
        @sd.expects(:process_generic).with(expected_cert, teardown_hash, 'apply', true)

        thread = @td.process_server_teardown_component(component).join

        raise thread[:exception] if thread[:exception]
      end
    end

    describe "when doing storage teardowns" do
      # TODO: Test compellent and netapp teardowns but need stub components
      it "should clean up compellent storage volumes"

      it "should attempt to remove the storage from the cluster" do
        component = @components.select{|c| c["type"] == "STORAGE"}.first
        resource_hash = ASM::Util.build_component_configuration(component, :decrypt => true)

        @td.expects(:remove_storage_component_from_cluster).with(component, resource_hash)
        @td.expects(:cleanup_compellent_storage_volumes).returns(true)
        @sd.expects(:process_generic).with("equallogic-eql-env02", resource_hash, "device", true, nil, "equallogic-172.17.2.10")

        thread = @td.process_storage_teardown_component(component).join
        raise thread[:exception] if thread[:exception]
      end

      it "should attempt to clean up compellent volumes using a specific method" do
        component = @components.select{|c| c["type"] == "STORAGE"}.first
        resource_hash = ASM::Util.build_component_configuration(component, :decrypt => true)

        @td.expects(:remove_storage_component_from_cluster)
        @td.expects(:cleanup_compellent_storage_volumes).with(component, resource_hash).returns(false)
        @sd.expects(:process_generic).never

        thread = @td.process_storage_teardown_component(component).join
        raise thread[:exception] if thread[:exception]
      end

      it "should remove equalogic storage components from the cluster" do
        component = @components.select{|c| c["type"] == "STORAGE"}.first
        server_component = @components.select{|c| c["type"] == "SERVER"}
        cluster_component = @components.select{|c| c["type"] == "CLUSTER"}

        resource_hash = ASM::Util.build_component_configuration(component, :decrypt => true)

        @sd.expects(:find_related_components).with("SERVER", component).returns(server_component)
        @sd.expects(:find_related_components).with("CLUSTER", server_component.first).returns(cluster_component)
        ASM::Util.expects(:get_eql_volume_iqn).returns("RSPECIQN")

        resources = {"transport" => {"vcenter" => {"name" => "vcenter-env02-vcenter.aidev.com",
                                                   "options" => {"insecure" => true},
                                                   "provider" => "device_file"}},
                     "esx_datastore" => {"172.28.2.125:arriTest" => {"ensure" => "absent",
                                                                     "datastore" => "arriTest",
                                                                     "type" => "vmfs",
                                                                     "target_iqn" => "RSPECIQN",
                                                                     "transport" => "Transport[vcenter]"}}}

        @sd.expects(:process_generic).with("equallogic-eql-env02", resources, "apply", true, nil, "equallogic-172.17.2.10")

        @td.remove_storage_component_from_cluster(component, resource_hash)
      end

      # need stubs
      it "should remove compellent storage components from the cluster"
      it "should remove netapp storage components from the cluster"
    end

    describe "when doing cluster teardowns" do
      it "should check the component has a certname" do
        thread = @td.process_cluster_teardown_component({}).join
        thread[:exception].message.should == "Component has no certname"
      end

      it "should clean the resource hash and call process_generic" do
        component = @components.select{|c| c["type"] == "CLUSTER"}.first
        @sd.stubs(:find_related_components).with("VIRTUALMACHINE", component).returns([])
        @td.stubs(:cleanup_vmware_vm_razor_policy).returns(true)
        @sd.expects(:find_related_components).with("SERVER", component).returns([])
        cleaned_resource_hash = {"asm::cluster" =>
                                  {"vcenter-env02-vcenter.aidev.com" =>
                                    {"datacenter" => "arri_vdc",
                                     "cluster" => "arri_cluster",
                                     "ensure" => "absent",
                                     "vcenter_options" => {"insecure" => true}}}}
        @sd.expects(:process_generic).with('vcenter-env02-vcenter.aidev.com', cleaned_resource_hash, "apply", true)
        thread = @td.process_cluster_teardown_component(component).join
        raise thread[:exception] if thread[:exception]
      end

      it "should call process_generic on related servers" do
        cluster_component = @components.select{|c| c["type"] == "CLUSTER"}.first
        server_component = @components.select{|c| c["type"] == "SERVER"}.first

        @sd.stubs(:find_related_components).with("VIRTUALMACHINE", cluster_component).returns([])
        @td.stubs(:cleanup_vmware_vm_razor_policy).returns(true)
        @sd.expects(:find_related_components).with("SERVER", cluster_component).returns([server_component])
        @sd.expects(:lookup_hostname).returns('arri-testhost')
        cluster_hash = {"asm::cluster" =>
                                     {"vcenter-env02-vcenter.aidev.com" =>
                                          {"datacenter" => "arri_vdc",
                                           "cluster" => "arri_cluster",
                                           "ensure" => "absent",
                                           "vcenter_options" => {"insecure" => true}}}}
        server_hash = {'asm::host' => {},
                         'transport' => {
                             'vcenter' => {
                                 'name' => "vcenter-env02-vcenter.aidev.com",
                                 'options' => {'insecure' => true},
                                 'provider' => 'device_file'}}}
        server_hash['asm::host']['bladeserver-jgd5hw1'] = {
            'datacenter' => 'arri_vdc',
            'cluster' => 'arri_cluster',
            'hostname' => 'arri-testhost',
            'username' => 'root',
            'password' => 'ff80808148a06b000148a6d3e6f70acf',
            'decrypt' => true,
            'timeout' => 90,
            'ensure' => 'absent'
        }
        @sd.expects(:process_generic).with('vcenter-env02-vcenter.aidev.com', cluster_hash, "apply", true)
        @sd.expects(:process_generic).with('bladeserver-jgd5hw1', server_hash, "apply", true)
        thread = @td.process_cluster_teardown_component(cluster_component).join
        raise thread[:exception] if thread[:exception]
      end
    end

    describe 'when doing vm teardowns' do
      it 'should not have expected parameters' do
        component = @components.select{|c| c["type"] == "VIRTUALMACHINE"}.first
        cluster = @components.select{|c| c["type"] == "CLUSTER"}.first
        ASM::DeviceManagement.stubs(:parse_device_config)
            .with('vcenter-env02-vcenter.aidev.com')
            .returns('rspec-vcenter-conf')
        ASM::Util.stubs(:find_vm_uuid).with('rspec-vcenter-conf', 'arri-testvm2')
            .returns('rspe-c288-0e73-104e-e6c0-31f5f65ad063')
        @sd.expects(:find_related_components).with('CLUSTER', component).returns([cluster])
        @sd.stubs(:find_related_components).with("VIRTUALMACHINE", component).returns([])
        @td.stubs(:cleanup_vmware_vm_razor_policy).returns(true)
        resource_hash = {
            'asm::vm::vcenter' => {
                'arri-testvm2' => {
                    'cpu_count' => '1',
                    'disksize_in_gb' => '32',
                    'memory_in_mb' => '8192',
                    'ensure' => 'absent',
                    'os_type' => 'linux',
                    'os_guest_id' => 'rhel6_64Guest',
                    'scsi_controller_type' => 'VMware Paravirtual',
                    'cluster' => 'arri_cluster',
                    'datacenter' => 'arri_vdc',
                    'vcenter_id' => 'vcenter-env02-vcenter.aidev.com',
                    'vcenter_options' => {'insecure' => true}
                },
            },
            'asm::server' => {
                'arri-testvm2' => {
                    'os_host_name' => 'arri-testvm2',
                    'admin_password' => 'ff80808148a06b000148a6d3e6d20acb',
                    'decrypt' => true,
                    'razor_image' => 'rhel6',
                    'ensure' => 'absent',
                    'broker_type' => 'noop',
                    'serial_number' => '',
                    'policy_name' => 'policy-arri-testvm2-8888',
                    'installer_options' => {'os_type' => 'redhat', 'agent_certname' => 'agent-arri-testvm2'},
                },
            },
        }
        @sd.expects(:process_generic).with('vm-arri-testvm2', resource_hash, 'apply')
        thread = @td.process_virtualmachine_teardown_component(component).join
        raise thread[:exception] if thread[:exception]
      end
    end

    describe "when doing generic teardowns" do
      it "should check component has a certname" do
        thread = @td.process_teardown_component({}).join
        thread[:exception].message.should == "Component has no certname"
      end

      it "should use process_generic to do the teardown" do
        component = @components.select{|c| c["type"] == "VIRTUALMACHINE"}.first

        ASM::Util.expects(:build_component_configuration).returns([])
        @sd.expects(:process_generic).with("FE8E1A17-259B-427A-ADA6-A2581F19B012", [], "apply", true)

        thread = @td.process_teardown_component(component).join

        raise thread[:exception] if thread[:exception]
      end
    end
  end
end
